c     Molscat utility program to combine several ISAVEU files into one
c     big sorted file for later processing, by G. C. McBane,
c     mcbaneg@gvsu.edu.  The main Molscat page is 
c     http://www.giss.nasa.gov/tools/molscat/

c     This program is part of the PMP Molscat package.  The most recent
c     version should be available at
c     http://faculty.gvsu.edu/mcbaneg/pmpmolscat/index.html

c     It will work with files in the version 14 ISAVEU format, whether
c     they were generated by PMP Molscat, the serial Molscat version 14,
c     or any other program.

c     NAMELIST input variables are required to specify
c     the number of input files, their names, and the name
c     of the output file.  Logical variable 'autoname' can
c     be set to .true. in the Namelist block to make smerge
c     generate input filenames ISAVEU.0000, ISAVEU.0001,
c     etc., as generated by PMP Molscat.  Integer variables
c     jtotmin and jtotmax may be used to suppress the 
c     inclusion of S-matrices with JTOT outside that range
c     in the merged output file.

c     An example input file, which will combine the three
c     files ISAVEU.0000, ISAVEU.0001, and ISAVEU.0002 into
c     one large file called totaldata.smat:

c &merge
c      numfiles = 3
c      autoname = .true.
c      outputfilename = 'totaldata.smat'
c /

c     Another example, combining a.smat and b.smat into out.smat,
c     and including only S-matrices with JTOT <= 20 in out.smat:

c &merge
c      numfiles = 2
c      jtotmax = 20
c      filenames(1) = 'a.smat'
c      filenames(2) = 'b.smat'
c      outputfilename = 'out.smat'
c /

c     With some Fortran compilers, both the & and / characters
c     must be in column 2 of the input file.

c     Internal storage for S-matrix data is defined in sdata.fi;
c     storage can be increased for large jobs.

c     The temporary files this program produces are legitimate
c     ISAVEU files in themselves, and can be used as input
c     files if a run terminates prematurely.
c     

c     The program is wildly non-reentrant; it must be
c     started anew for each group of ISAVEU files to be combined.


c     Revision history:

c     Begun 19 January 2005
c     First version working 29 January 2005

c     31 January 2005: modified to try to better handle truncated
c     ISAVEU files.  

c     2 February 2005: added 'autoname' option to NAMELIST
c     block to permit automatic generation of filenames
c     ISAVEU.0000, ISAVEU.0001,...  as produced by PMP Molscat.

c     6-10 March 2005: 
c     Returned to original 'smerge' name
c     Modified to go on to next input file, rather than halting,
c        in many more cases, in particular when first input file 
c        is empty (typical of dynamic version of PMP Molscat)
c     Rewrote and cleaned up read_sdata
c     Modified to skip duplicate S-matrices; will write only
c        the first one with a particular key.  This should make it
c        easier to use when failed Molscat runs are restarted, since
c        that will often produce a few duplicated S-matrices.
c     Modified to use plain F77 mechanism for deleting temporary
c        files, rather than nice Compaq Fortran extension

c     March 11-19 2005:
c     Fixed bug introduced in rewrite of read_sdata (duh)
c     Fixed more subtle bug that only showed up when data files
c     contained an S-matrix with more than 800 open channels
c     (had failed to check against available array size)

c     April 2, 2005:
c     Added code to handle S-matrix files with different ENERGY 
c     arrays.  Energies not encountered before are simply tacked
c     onto the end of the ENERGY array, NNRG is incremented,
c     and each new S-matrix read has its value of INRG adjusted
c     if necessary to match the global ENERGY array.  This permits
c     merging of data from runs that don't each do all
c     the energies of interest.  The design is a little clunky
c     (the eventual ordering of the ENERGY values depends on
c     the order in which S-matrix files are listed in the
c     NAMELIST &MERGE input, and later ones do more searching around
c     in the ENERGY array) but the overhead is small.
c     There is a new parameter echktol in gheader.fi that sets
c     how close together energy values have to be for the program
c     to consider them equal.

c     April 19--20, 2005:
c     Modified to provide sensible behavior when maximum number
c     of temporary files is exceeded.  Default value of mxfdata in
c     sdata.fi increased to reduce likelihood of users encountering
c     this condition in the first place.

c     September 28, 2005:
c     Added Namelist variables jtotmin and jtotmax, to permit
c     suppression of JTOT values outside their range in the merged
c     file.

c     November 12, 2007:
c     Modified to provide useful error message when amount of
c     storage for integer data (idata/mxidata) is inadequate
c     for a single S-matrix.  (Behavior was already sensible
c     if the floating point storage was inadequate.)
c     Modified comments in sdata.fi to give better information
c     about required storage for a given size S-matrix.
c     Changed temporary file names to use 3-digit suffix
c     rather than 2-digit, so more than 99 temporary
c     files can be accomodated.


c     Portability issues: 

c     The program is mostly straight Fortran 77, except:
c     Long variable and routine names are used.
c     Namelist input is used.
c     The DO WHILE, CYCLE and EXIT statements are used.

c     All the F77 compilers I know of support those extensions, though
c     there are some annoying variations in what kind of formatting
c     is acceptable in NAMELIST input.

c     Known bugs:
      
c     No attempt is made to handle ISAVEU files generated by Molscat
c        versions before v14.
c     Some error exits are missing; the program may crash on disk 
c        full, etc.
c     No advantage is taken of the common case where entire data set fits
c        into memory. Possible savings in disk usage are not gained.
c        This would be easy if Fortran had a RENAME statement.


      program smerge
      implicit none

c     variables needed for filename list
      include 'inputfiles.fi'

c     instruction definitions for read_sdata
      include 'elementtypes.fi'

c     information about temporary files
      include 'tempfiles.fi'

      character*80 outputfilename
      integer inputunit, jtotmin, jtotmax
      logical autoname   ! whether we generate names of ISAVE files 
      namelist /merge/ numfiles, filenames, outputfilename,
     1     inputunit, autoname, jtotmin, jtotmax

      integer outputunit, tbaseunit
      parameter(outputunit = 4, tbaseunit=10)

c     "small" data for each S-matrix ("big" data in common /sdata/)
      integer mxsmat
      parameter(mxsmat=10000)
      integer index(7, mxsmat), keys(mxsmat), iptr(mxsmat)
c     Each column of index() describes a stored S-matrix.  The
c     element values for matrix number k are as follows:
c     index(1,k) = jtot
c     index(2,k) = inrg
c     index(3,k) = iexch
c     index(4,k) = m
c     index(5,k) = nopen
c     index(6,k) = element of fdata array where floating-point values 
c        begin;  starting at fdata(index(6,k)) we find energy, wt, 
c        (wvec(i), i = 1, nopen), ((sreal(j,i), j=1,i), i=1,nopen),
c        ((simag(j,i), j=1,i), i=1,nopen) 
c     index(7,k) = element of idata array where integer values begin;
c        starting at idata(index(7,k)) we find 
c        (j(i), i=1,nopen), (l(i),i=1,nopen)
      

c     local variables
      integer i, idum, n, numtempfiles, num_done, oldkey
      double precision fdum
      logical allread, memfull, success
      integer IBIG
      parameter (IBIG=2**30)  !larger than any reasonable key

      double precision echkwt(2,maxtempfiles)

c     functions
      integer makekey, get_smatrix, select_next

c     variable for communication with ISORT routine
      integer kflag
      parameter(kflag = 2)


c     set default values for namelist read
      outputfilename = 'sortmerge.smat'
      numfiles = 1
      filenames(1) = 'ISAVEU.0000'
      inputunit = 8
      autoname = .false.
      jtotmin = 0
      jtotmax = IBIG

c     get list of files to process
      read(*,merge)

      if (numfiles.gt.maxmergefiles) then
         write(*, 500) numfiles, maxmergefiles
         stop
      end if
  500  format(1x,'numfiles (', I3, '), is greater than maxmergefiles (',
     1     I2,').',/,
     2     ' Increase maxmergefiles in smerge.f and recompile.')

       If (autoname) call namegen

      allread = .false.
      numtempfiles = 0

c     read in global header from first file.  All arguments beyond
c     the first two are ignored by read_sdata on this call.
      call read_sdata(inputunit, global_header, idum, fdum, idum, 
     1     allread)

c     Read in as many S-matrices as we can from input files.  When
c     they are all read or the memory is full, sort them and
c     write them out in a temporary file. Repeat until all the 
c     S-matrices have been read and written to sorted temporary files.

      !loop over temporary files
      do while ((.not. allread).and.(numtempfiles.lt.maxtempfiles))  
         call clear_sdata(memfull)
         n = 0       ! number of S-matrices read so far 
         do while ((.not. allread) .and. (.not. memfull) .and. 
     1        (n.lt.mxsmat) )     !loop over S-matrices      
            if (get_smatrix(inputunit, index(1,n+1), 
     1           allread, memfull)   .eq. 1) then
                         !opens and closes input files as necessary
                         !returns 1 if S-matrix fully read, 0 if not
               n = n+1   ! update count of S-matrices in memory
               iptr(n) = n  ! insert pointer to column in index()
               keys(n) = makekey(index(1,n), index(4,n), index(2,n))
            end if
         end do
         call isort(keys, iptr, n, kflag) !on three-way key: JTOT|M|INRG
         call write_tempfile(iptr, index, n, numtempfiles)
         numtempfiles = numtempfiles+1
      end do

      if (.not.allread) then
         write(*,900) 
      else
         write(*,*) ' All input files read.'
      end if
 900  format(1x, 'Maximum number of temporary files exceeded.',/,
     1     ' Increase mxfdata and/or mxidata in sdata.fi, or',/,
     2     ' increase maxtempfiles in tempfiles.fi, and recompile.',/,
     3     ' Execution continuing with current set of temporary files.')

c     Now merge those sorted temporary files into one final
c     sorted output file. 

c     write main header for output file; if this fails, temp files retained
      open(outputunit, file = outputfilename, form='unformatted', 
     1     status = 'new', err = 1200)
      call write_global_header(outputunit)

c     open all temporary files for multiway merge
      do i = 1, numtempfiles
         open(tbaseunit+i, file = tempfilenames(i), form='unformatted')
         call check_gheader(tbaseunit+i, success)
      end do

      numfiles = 1
      allread = .false.
      memfull = .false.
c     reset read_sdata to make it do one file at a time
c     only the "reset" argument matters; others not used
      call read_sdata(idum, reset, idum, fdum, idum, allread)

c     load info on first S-matrix in each file 
      do i = 1, numtempfiles
         call read_sdata(tbaseunit+i, s_header, index(1,i), echkwt(1,i),
     1      idum, allread)
         keys(i) = makekey(index(1,i), index(4,i), index(2,i))
      end do

c     merge
      num_done = 0
      allread = .false.
      oldkey = -1        ! can't match any real key
      do while (num_done .lt. numtempfiles)
         i = select_next(keys, numtempfiles)  !pick next S-matrix
         if ( (keys(i) .ne. oldkey) .and. 
     1        (index(1,i).ge.jtotmin) .and. ! index(1,i) is jtot(i)
     2        (index(1,i).le.jtotmax) ) then              
!     move it from temporary to permanent file
            call transfer_smatrix(tbaseunit+i, outputunit, index(1,i), 
     1           echkwt(1,i), .true.)   
            oldkey = keys(i)
            write(*,*) 'writing S-matrix: JTOT = ', index(1,i), 
     1           'M = ', index(4,i), 'INRG = ', index(2,i)
         else  !duplicate S-matrix or out of j-range; discard it
            call transfer_smatrix(tbaseunit+i, outputunit, index(1,i), 
     1           echkwt(1,i), .false.)
         end if
         call read_sdata(tbaseunit+i, s_header, index(1,i), echkwt(1,i),
     1      idum, allread) ! get next S-header from that same temp file
         if (allread) then !if that one is empty, clean up
            num_done = num_done+1
            keys(i) = IBIG  !ensure it won't be selected any more
            allread = .false.
         else   !otherwise compute its key
            keys(i) =  makekey(index(1,i), index(4,i), index(2,i))
         end if
      end do
      close(outputunit)

c     Delete temporary files
      do i = 1, numtempfiles
         open(tbaseunit, file = tempfilenames(i), form='unformatted')
         close(tbaseunit, status = 'delete')
      end do

      write(*,1100) outputfilename
 1100 format(1x, 'Merge completed; data written to ', A30)
      stop   !normal termination

c     Output file probably already there; don't overwrite
c     (it might have taken months to obtain!)
 1200 write(*,1210) outputfilename
      stop
 1210 format(1x, 'Error opening output file ', A20, '.',/,
     1     ' Does it exist already?', /, 
     2     ' Data from this run retained in files smtemp.nnnn.',/,
     3     ' Bailing out.')
      
      end

c     Get a single S-matrix from the input files.  If there
c     is insufficient memory to store it, set memfull and 
c     return 0, saving its header for use later.  If there are no
c     more matrices to read, set allread and return 0.  Otherwise,
c     read S-matrix and associated data into memory, 
c     and return 1.
      integer function get_smatrix(inputunit, index, 
     1     allread, memfull)  
      implicit none
      integer inputunit, index(7)
      logical allread, memfull

      include 'elementtypes.fi'
      include 'sdata.fi'

      integer idum, fneeded, ineeded, nopen, i
      integer locindex(7)
      double precision locfdata(2)
      save locindex, locfdata

      logical waiting, done
      data waiting /.false./
      save waiting
      
      done = .false.
      do while (.not. done)

c     Get an S-matrix header, one way or another
         if (.not. waiting) then  !must read one off disk
            call read_sdata(inputunit, s_header, locindex, 
     1           locfdata, idum, allread)
            if (allread) then
               done = .true.
               get_smatrix = 0
               return
            end if
         end if
     
c     check for memory full condition, either temporary or fatal
         nopen = locindex(5)
         fneeded = 2 + nopen + nopen*(nopen+1) !wvec + S storage
         ineeded = 2*nopen
         if ( (fneeded.gt.(mxfdata-nextfpos+1)) .or. 
     1        (ineeded.gt.(mxidata-nextipos+1)) ) then  !memory full
            if (fneeded.gt.mxfdata) then   !no hope; bail out
               write(*,500) fneeded, mxfdata
               stop
 500           format(1x, ' One S-matrix requires ', I20, 
     1              ' words of storage; mxfdata = ', I20, '.',/,
     2              ' Increase mxfdata in sdata.fi and recompile.')
            else if (ineeded.gt.mxidata) then   !no hope; bail out
               write(*,510) ineeded, mxidata
               stop
 510           format(1x, ' One S-matrix requires ', I20, 
     1              ' words of integer storage; mxidata = ', I20, '.',/,
     2              ' Increase mxidata in sdata.fi and recompile.')
            else  ! report back mem full; expect to try again
               memfull = .true.
               waiting = .true.
               done = .true.
               get_smatrix = 0
            end if

         else  ! memory not full; load S-matrix data

            do i = 1, 5
               index(i) = locindex(i)
            end do
            index(6) = nextfpos
            index(7) = nextipos
            fdata(nextfpos) = locfdata(1)
            fdata(nextfpos+1) = locfdata(2)

            call read_sdata(inputunit, s_data, index, 
     1           fdata(nextfpos+2), idata(nextipos), allread)
            if (index(1).ge.0) then ! everything ok; return data to caller
               nextfpos = nextfpos+fneeded
               nextipos = nextipos+ineeded
               waiting = .false.
               done = .true.
               get_smatrix = 1
            else                ! unexpected EOF on data read; 
                                ! if there are more files, go around again
               waiting = .false.
               get_smatrix = 0
               if (.not. allread) then
                  done = .false.
               else
                  done = .true.
               end if
            end if
         end if
      end do
      return
      end



      subroutine read_sdata(inputunit, element, index, faddr, iaddr, 
     1     allread)
c     read the next S-matrix entry
c     element can be global_header, s_header, s_data, or reset
c     allread used to signal completion of last input file
c     index(1) =-1 signals unexpected EOF if element = s_data

      implicit none
      integer inputunit, element,  iaddr(*)
      double precision faddr(*)
      logical allread
      integer index(7)

      include 'elementtypes.fi'  !definitions for 'element'
      include 'inputfiles.fi'  ! list of S-matrix input files
      include 'gheader.fi'     ! global header data
      integer nsq, i

      integer mxch
      parameter(mxch = 2000)
      integer j(mxch), l(mxch)
      double precision wvec(mxch)

      integer jtot,inrg,iexch,m,nopen , iend 
      double precision echk, wt
      logical incomplete

      integer currentfile    ! keep up with which input file
      data currentfile /0/
      save currentfile

      allread = .false.

      if (element .eq. reset) then
         currentfile = 1

      else if (element .eq. global_header) then
         if(currentfile.eq.0) then !  first call; open first file
            call next_file(inputunit, element, currentfile, allread)
         end if

         do while (.not.allread) 
            READ(inputunit, err=600) LABEL,ITYPE,NLEV,NQN,URED,IP
            NSQ=NLEV*NQN
            IF (NSQ.GT.MXJLEV) THEN
               WRITE(6,*) ' *** NLEV*NQN.GT.MXJLEV',MXJLEV
               STOP
            ENDIF
            READ(inputunit, err=600) (JLEV(I),I=1,NSQ)
            READ(inputunit, err=600) NLEVEL,(ELEVEL(I),I=1,NLEVEL)
            READ(inputunit, err=600) NNRG,(ENERGY(I),I=1,NNRG)
            exit   !successful read of global header
            !here on bad read; continue to next file
 600        call next_file(inputunit, element, currentfile, allread)
         end do

      else if (element .eq. s_header) then ! first line of an S-matrix section
         do while (.not. allread)
            READ(inputunit, end=650, err=650) JTOT,INRG,ECHK,IEXCH,WT,
     1        M,NOPEN  
            if (echk.ne.energy(inrg)) then  !mismatched energy
               call new_energy(inrg, echk)  ! inrg will be changed
            end if                          ! to match ENERGY array           
            exit
 650        call next_file(inputunit, element, currentfile, allread)
         end do 

         index(1) = jtot
         index(2) = inrg
         index(3) = iexch
         index(4) = m
         index(5) = nopen
         faddr(1) = echk
         faddr(2) = wt

      else if (element .eq. s_data) then ! data from S-matrix section
         nopen = index(5)
         if (nopen.gt.mxch) then
            write(*, 1100) nopen, mxch
            stop
         end if
 1100    format(1x, 'Too many open channels.  Increase mxch in ',
     1        'read_sdata and recompile.  nopen, mxch = ', 2(I5, 2x))

         READ(inputunit,err=670, end=670) (J(I),L(I),WVEC(I),I=1,NOPEN)
         ! copy vectors into indexed storage
         do i = 1, nopen
            iaddr(i) = J(i)
            iaddr(nopen+i) = L(i)
            faddr(i) = wvec(i)
         end do

         !now read S-matrix directly into storage
         CALL ltsread(inputunit,NOPEN,faddr(nopen+1),IEND)
         if (iend.ne.0) goto 670
         call ltsread(inputunit,NOPEN,faddr(nopen+nopen*(nopen+1)/2+1),
     1        IEND)
         if (iend.ne.0) goto 670
        
      else   !invalid element
         write(*,700) element
         stop
 700     format(1x, 'Invalid value of element: ', I3, '.', /, 
     1        ' Bailing out.') 

      end if

      return   !common return for successful operation

!     here on unexpected EOF on s_data read
 670     write(*,660) filenames(currentfile), index(1), index(4), 
     1        index(2)
 660     format(1x, 'Unexpected end of file encountered on file ', A20,
     1        /, ' at J, M, INRG = ', 3(I3, 2x), '.', /, 
     2        ' Continuing to next file.')

         index(1) = -1   ! signal that the data read failed
         call next_file(inputunit, element, currentfile, allread)
         ! depend on calling routine to catch problem 
         ! and call for another header
         return

         end

c     Subroutine to close current input file, open new one,
c     and position the file pointer to the appropriate spot

      subroutine next_file(inputunit, element, currentfile,
     1     allread)
      implicit none
      integer inputunit, element, currentfile
      logical allread

      logical successful

      include 'inputfiles.fi'
      include 'elementtypes.fi'

      allread = .false.
      successful = .false.
      do while (.not. (allread.or.successful)) 

         if (currentfile.ne.0) then
            close(inputunit)
         end if
         currentfile = currentfile+1
         if (currentfile .gt. numfiles)  then !we're done
            allread = .true.
         else
            open(inputunit, file = filenames(currentfile), 
     1           form = 'unformatted', status = 'old',  action = 'read', 
     2           err = 500)
            goto 510
 500        write(*,505) filenames(currentfile) !here for failed open
            cycle               ! print notice and try again
 505        format(1x, 'next_file: failed to open file ', A20, '.', /, 
     1           'Continuing.')
 510        continue            ! here if file apparently opened correctly
            if (element.ne.global_header) then
               call check_gheader(inputunit, successful)
            else
               successful = .true.
            end if
         end if
      end do
      return
      end
      
      
      

      subroutine ltsread(IU,N,S,IEND)
c     different from Molscat SREAD because it does not fill
c     in the upper half of the S matrix; retains lower triangular
c     packed storage
      implicit none
      integer IU, N, IEND
c      double precision S(n,n)
      double precision S(*)
      integer i,j

      IEND=0
c      READ(IU,END=9999) ((S(I,J),J=1,I),I=1,N)
      READ(IU,END=9999) (S(I),I=1,(N*(N+1))/2)
      RETURN
 9999 IEND=1
      RETURN
      END

      subroutine ltswrite(IU,N,S)
c     different from Molscat SWRITE because it
c     assumes lower triangular packed storage as used in ltsread
      implicit none
      integer IU, N
c      double precision S(n,n)
      double precision S(*)
      integer i,j

c      write(IU,err=9999) ((S(I,J),J=1,I),I=1,N)
      write(IU,err=9999) (S(I),I=1,(N*(N+1))/2)
      RETURN
 9999 write(*,*) ' Error writing S-matrix.  Disk full?'
      stop
      END

      subroutine check_gheader(inputunit, success)
!     Called on opening of ISAVEU files after the first.  Checks to
!     make sure information in the header is compatible with that
!     from the first file that was opened, and positions file
!     pointer to beginning of S-matrix sections for this new file.

!     This version halts if a serious mismatch is found; it might also
!     be reasonable to continue to the next input file.  Replacing
!     STOP with GOTO 800 in this function would accomplish that,
!     though the error messages would be confusing.



      implicit none
      integer inputunit
      logical success

      include 'gheader.fi'
      CHARACTER*4 locLABEL(20)
      integer locitype, locnlev, locnqn, locip, locnlevel, locnnrg
      integer locjlev(MXJLEV), nsq, i
      double precision locured, locelevel(MXEL), locenergy(MXNRG)
      double precision uredmismatch

      success = .true.

      READ(inputunit, err=800) loclabel,locitype,locnlev,locnqn,
     1     locured,locip

      if ((locitype.ne.itype).or.(locnlev.ne.nlev).or.
     1     (locnqn.ne.nqn).or.(locip.ne.ip)) then
         write(*,*) 'Integer parameter mismatch'
         stop
      end if
         
      uredmismatch = abs(locured-ured)
      if (uredmismatch.gt.1.0d-4) then
         if (uredmismatch.gt.0.1d0) then
            write(*,*) 'Mismatch in reduced masses.  Bailing out.'
            stop
         else
            write(*,*) 'Mismatch in reduced masses.  Continuing.'
         end if
      end if

      NSQ=NLEV*NQN
      READ(inputunit, err=800) (locJLEV(I),I=1,NSQ)
      READ(inputunit, err=800) locNLEVEL,(locELEVEL(I),I=1,locNLEVEL)
      READ(inputunit, err=800) locNNRG,(locENERGY(I),I=1,locNNRG)
      
      do i = 1, nsq
         if (locjlev(i).ne.jlev(i)) then
            write(*,500)  i, locjlev(i), jlev(i)
            stop
 500        format(1x, 'JLEV(', I3, ') = ', I3, 'does not match',
     1           'original value of ', I3, '.',/,' Bailing out.'  )
         end if
      end do
      do i = 1, nlevel
         if (abs(locelevel(i)-elevel(i)).gt.1.0d-4) then
            write(*,600)  i, locelevel(i), elevel(i)
            stop
 600        format(1x, 'ELEVEL(', I3, ') = ', G10.4, 
     1           'does not match original value of ', G10.4, '.',/,
     2           ' Bailing out.'  )
         end if
      end do
      
c     Look through energy list for surprising values; if we find
c     any, call extend_elist to add them to the overall list if
c     necessary.

      do i = 1, locnnrg
         if (abs(locenergy(i)-energy(i)).gt.echktol) then
            call extend_elist(locenergy(i))
         end if
      end do

      return    ! successful completion

c     Arrive here if one of the read statements fails
 800  write(*, 810) inputunit   
      success = .false.
      return

 810  format(1x,' check_gheader: error reading from input unit ',
     1     I2,'.',/,' Continuing to next file.')
      end

      subroutine clear_sdata(memfull)
      implicit none
      include 'sdata.fi'
      logical memfull
      
      nextfpos = 1
      nextipos = 1
      memfull = .false.
      return
      end

c     This function is called if a newly read S-matrix header has an
c     ECHK that does not match the global header's energy list.  This
c     will happen if several runs with different NNRG and ENERGY()
c     arrays are combined.  Check to see if the desired energy is
c     already in the energy list.  If it is, change INRG to the
c     appropriate value and return.  If it isn't, we have a problem; it
c     should have been put there by extend_elist when the current global
c     header was read.

      subroutine new_energy(inrg, echk)
      implicit none
      integer inrg
      double precision echk

      integer i

      include 'gheader.fi'

      do i = 1, nnrg
         if ( abs(echk-energy(i)) .lt. echktol) then  !found the right place
            inrg = i
            return   
         end if
      end do

      write(*,100) echk, inrg 
      stop
 100  format(1x, ' Error: mismatch between ECHK and energy list.', /, 
     1     ' ECHK = ', F10.4, ' INRG = ', I4, '.',/,
     2     ' Bailing out.')
      end

!  This function is called from check_gheader if a newly read global
c     header has an energy in its energy list that does not match the
c     one in the same position in the existing energy list (that is, the
c     ENERGY() array in COMMON /gheader/).  We check to see if the "new"
c     energy is already there in some other element of the energy list.
c     If it is, no problem; new_energy() will deal with the results.  If
c     it isn't, then try to extend the ENERGY array to accomodate the new
c     energy.

      subroutine extend_elist(newenergy)
      implicit none
      double precision newenergy

      integer i

      include 'gheader.fi'

      do i = 1, nnrg
         if ( abs(newenergy-energy(i)) .lt. echktol) then  ! it's there
            return
         end if
      end do

      if (nnrg .eq. mxnrg) then   !no room to extend list; fail
         write(*,100) 
         stop
      end if
 100  format(' Too many energies.  Stick with S-matrix files',/,
     1        ' that COULD be generated in a single Molscat run.',/,
     2        ' The usual maximum is 100 energies.  Bailing out.')

      nnrg = nnrg+1
      energy(nnrg) = newenergy

      return
      end

c     generate single compound key for sorting on jtot, m, inrg
c     assumes inrg.le.100 and m.le.1000
      integer function makekey(jtot, m, inrg)
      implicit none
      integer jtot, m, inrg

      if ((inrg.gt.100).or.(m.gt.1000)) then
         write(*, 100) inrg, m
 100     format(1x, 'Warning: key generation routine not designed',
     1        ' for inrg>100 or m>1000.  Sorting results  suspect.')
      end if

      makekey = inrg+100*m+100000*jtot
      return
      end

c     Read an S-matrix from temporary file.  If action is true,
c     also write it to permanent file.
      subroutine transfer_smatrix(inputunit, outputunit, index,
     1     echkwt, action)
      implicit none
      integer inputunit, outputunit, index(7)
      double precision echkwt(2)
      logical action

      include 'sdata.fi'
      include 'elementtypes.fi'

      logical ldum

c     use beginning of storage area 
      index(6) = 1
      index(7) = 1

c     put echk and wt in first two positions
      fdata(1) = echkwt(1)
      fdata(2) = echkwt(2)

      call read_sdata(inputunit, s_data, index, 
     1        fdata(3), idata(1), ldum)

      if (action) then
         call write_smatrix(outputunit, index)
      end if

      return
      end

      subroutine write_tempfile(iptr, index, n, filenum)
c     memory is full of n S-matrices; write out a file containing them,
c     in order indicated by iptr array
      implicit none
      integer iptr(*), index(7,*), n, filenum

      include 'tempfiles.fi'

      integer tempunit
      parameter(tempunit = 3)

      character*7 namestub
      parameter(namestub = 'smtemp.')

      character*3 suffix
      integer i

      write( suffix, '(i3.3)' ) filenum
      tempfilenames(filenum+1) = namestub//suffix

      open(tempunit, file=tempfilenames(filenum+1), form='unformatted')

      call write_global_header(tempunit)
      
      do i = 1, n
         call write_smatrix(tempunit, index(1,iptr(i)))
      end do

      close(tempunit)
      write(*,100) n, tempfilenames(filenum+1)
 100  format(1x, 'Wrote ', I4, ' S-matrices to temporary file ',
     1     A30)

      return
      end

      subroutine write_global_header(tempunit)
c     simpleminded no-safety-checking version
      implicit none
      integer tempunit

      include 'gheader.fi'

      integer nsq, i

      nsq = nlev*nqn
      write(tempunit) LABEL,ITYPE,NLEV,NQN,URED,IP
      write(tempunit) (JLEV(I),I=1,NSQ)
      write(tempunit) NLEVEL,(ELEVEL(I),I=1,NLEVEL)
      write(tempunit) NNRG,(ENERGY(I),I=1,NNRG)

      return
      end

      subroutine write_smatrix(tempunit, index)
      implicit none
      integer tempunit
      integer index(7)

      include 'sdata.fi'

      integer jtot, inrg, iexch, m, nopen, fpos, ipos, i
      double precision echk, wt

      jtot = index(1)
      inrg = index(2)
      iexch = index(3)
      m = index(4)
      nopen = index(5)

      fpos = index(6)
      ipos = index(7)

      echk = fdata(fpos)
      wt = fdata(fpos+1)

      write(tempunit) jtot, inrg, echk, iexch, wt, m, nopen

      call line6write(tempunit, idata(ipos), idata(ipos+nopen),
     1     fdata(fpos+2), nopen)

      call ltswrite(tempunit, nopen, fdata(fpos+2+nopen))
      call ltswrite(tempunit, nopen, 
     1     fdata(fpos+2+nopen+nopen*(nopen+1)/2))

      return
      end

      subroutine line6write(tempunit, j, l, wvec, nopen)
      implicit none
      integer tempunit, nopen, j(nopen), l(nopen)
      double precision wvec(nopen)

      integer i

      write(tempunit) (J(I),L(I),WVEC(I),I=1,NOPEN)

      return
      end

c     Simple linear selection of lowest key
c     More sophisticated algorithm probably not worthwhile
      integer function select_next(keys, numtempfiles)
      implicit none
      integer numtempfiles
      integer keys(numtempfiles)

      integer i

      select_next = 1
      do i = 2, numtempfiles
         if (keys(i).lt. keys(select_next)) select_next = i
      end do
      return
      end

c     generate list of ISAVEU files produced by pmp Molscat
c     for numfiles processes
      subroutine namegen
      implicit none
      include 'inputfiles.fi'

c     local variables
      integer i

c     functions
      character*4 suffix   

      do i = 1, numfiles
         filenames(i) = 'ISAVEU.'//suffix(i-1)
      end do

      return
      end

*  Convert process ID (rank) to character *4 suffix for filenames
*     copied from Ian Bush's labDA function
******************************************************************************
******************************************************************************
*
      character*4 function suffix( number )
*
*  Converts the integer NUMBER to a four character string, left justified.
*
      integer number
*
*  Use internal file to acheive this.
*
      write( suffix, '(i4.4)' ) number
      return
*
      end

         

c     routine to write out one complete S-matrix data set,
c     for debugging
      subroutine dump_data(index)
      implicit none
      integer index(7)
      include 'sdata.fi'

      integer i, nopen
      
      write(*,*) 'dump_data'
      write(*,*) ' index: '   
      write(*,*) (index(i), i = 1, 7)
      write(*,*) ' idata: '   
      nopen = index(5)
      write(*,*) (idata(index(7)+i-1), i = 1, 2*nopen)
      write(*,*) ' echk, wt:'
      write(*,*) fdata(index(6)), fdata(index(6)+1)
      write(*,*) ' wvec:'
      write(*,*) (fdata(index(6)+1+i), i = 1,nopen)
      write(*,*) ' S:'
      write(*,*) (fdata(index(6)+1+nopen+i), i = 1, nopen*(nopen+1))

      return
      end

c     The ISORT routine was taken from the SLATEC library
c     available at netlib.  Its error-handling code was 
c     modified to remove calles to the SLATEC error routines 
c     but it is otherwise stock.

*DECK ISORT
      SUBROUTINE ISORT (IX, IY, N, KFLAG)
C***BEGIN PROLOGUE  ISORT
C***PURPOSE  Sort an array and optionally make the same interchanges in
C            an auxiliary array.  The array may be sorted in increasing
C            or decreasing order.  A slightly modified QUICKSORT
C            algorithm is used.
C***LIBRARY   SLATEC
C***CATEGORY  N6A2A
C***TYPE      INTEGER (SSORT-S, DSORT-D, ISORT-I)
C***KEYWORDS  SINGLETON QUICKSORT, SORT, SORTING
C***AUTHOR  Jones, R. E., (SNLA)
C           Kahaner, D. K., (NBS)
C           Wisniewski, J. A., (SNLA)
C***DESCRIPTION
C
C   ISORT sorts array IX and optionally makes the same interchanges in
C   array IY.  The array IX may be sorted in increasing order or
C   decreasing order.  A slightly modified quicksort algorithm is used.
C
C   Description of Parameters
C      IX - integer array of values to be sorted
C      IY - integer array to be (optionally) carried along
C      N  - number of values in integer array IX to be sorted
C      KFLAG - control parameter
C            =  2  means sort IX in increasing order and carry IY along.
C            =  1  means sort IX in increasing order (ignoring IY)
C            = -1  means sort IX in decreasing order (ignoring IY)
C            = -2  means sort IX in decreasing order and carry IY along.
C
C***REFERENCES  R. C. Singleton, Algorithm 347, An efficient algorithm
C                 for sorting with minimal storage, Communications of
C                 the ACM, 12, 3 (1969), pp. 185-187.
C***ROUTINES CALLED  XERMSG
C***REVISION HISTORY  (YYMMDD)
C   761118  DATE WRITTEN
C   810801  Modified by David K. Kahaner.
C   890531  Changed all specific intrinsics to generic.  (WRB)
C   890831  Modified array declarations.  (WRB)
C   891009  Removed unreferenced statement labels.  (WRB)
C   891009  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   900315  CALLs to XERROR changed to CALLs to XERMSG.  (THJ)
C   901012  Declared all variables; changed X,Y to IX,IY. (M. McClain)
C   920501  Reformatted the REFERENCES section.  (DWL, WRB)
C   920519  Clarified error messages.  (DWL)
C   920801  Declarations section rebuilt and code restructured to use
C           IF-THEN-ELSE-ENDIF.  (RWC, WRB)
C***END PROLOGUE  ISORT
C     .. Scalar Arguments ..
      INTEGER KFLAG, N
C     .. Array Arguments ..
      INTEGER IX(*), IY(*)
C     .. Local Scalars ..
      REAL R
      INTEGER I, IJ, J, K, KK, L, M, NN, T, TT, TTY, TY
C     .. Local Arrays ..
      INTEGER IL(21), IU(21)
C     .. External Subroutines ..
c      EXTERNAL XERMSG
C     .. Intrinsic Functions ..
      INTRINSIC ABS, INT
C***FIRST EXECUTABLE STATEMENT  ISORT
      NN = N
      IF (NN .LT. 1) THEN
         write(*,*) 'The number of values to be sorted is not positive.'
         stop
      ENDIF
C
      KK = ABS(KFLAG)
      IF (KK.NE.1 .AND. KK.NE.2) THEN
         write(*,*) 'The sort control parameter, K, is invalid.'
         stop
      ENDIF
C
C     Alter array IX to get decreasing order if needed
C
      IF (KFLAG .LE. -1) THEN
         DO 10 I=1,NN
            IX(I) = -IX(I)
   10    CONTINUE
      ENDIF
C
      IF (KK .EQ. 2) GO TO 100
C
C     Sort IX only
C
      M = 1
      I = 1
      J = NN
      R = 0.375E0
C
   20 IF (I .EQ. J) GO TO 60
      IF (R .LE. 0.5898437E0) THEN
         R = R+3.90625E-2
      ELSE
         R = R-0.21875E0
      ENDIF
C
   30 K = I
C
C     Select a central element of the array and save it in location T
C
      IJ = I + INT((J-I)*R)
      T = IX(IJ)
C
C     If first element of array is greater than T, interchange with T
C
      IF (IX(I) .GT. T) THEN
         IX(IJ) = IX(I)
         IX(I) = T
         T = IX(IJ)
      ENDIF
      L = J
C
C     If last element of array is less than than T, interchange with T
C
      IF (IX(J) .LT. T) THEN
         IX(IJ) = IX(J)
         IX(J) = T
         T = IX(IJ)
C
C        If first element of array is greater than T, interchange with T
C
         IF (IX(I) .GT. T) THEN
            IX(IJ) = IX(I)
            IX(I) = T
            T = IX(IJ)
         ENDIF
      ENDIF
C
C     Find an element in the second half of the array which is smaller
C     than T
C
   40 L = L-1
      IF (IX(L) .GT. T) GO TO 40
C
C     Find an element in the first half of the array which is greater
C     than T
C
   50 K = K+1
      IF (IX(K) .LT. T) GO TO 50
C
C     Interchange these elements
C
      IF (K .LE. L) THEN
         TT = IX(L)
         IX(L) = IX(K)
         IX(K) = TT
         GO TO 40
      ENDIF
C
C     Save upper and lower subscripts of the array yet to be sorted
C
      IF (L-I .GT. J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 70
C
C     Begin again on another portion of the unsorted array
C
   60 M = M-1
      IF (M .EQ. 0) GO TO 190
      I = IL(M)
      J = IU(M)
C
   70 IF (J-I .GE. 1) GO TO 30
      IF (I .EQ. 1) GO TO 20
      I = I-1
C
   80 I = I+1
      IF (I .EQ. J) GO TO 60
      T = IX(I+1)
      IF (IX(I) .LE. T) GO TO 80
      K = I
C
   90 IX(K+1) = IX(K)
      K = K-1
      IF (T .LT. IX(K)) GO TO 90
      IX(K+1) = T
      GO TO 80
C
C     Sort IX and carry IY along
C
  100 M = 1
      I = 1
      J = NN
      R = 0.375E0
C
  110 IF (I .EQ. J) GO TO 150
      IF (R .LE. 0.5898437E0) THEN
         R = R+3.90625E-2
      ELSE
         R = R-0.21875E0
      ENDIF
C
  120 K = I
C
C     Select a central element of the array and save it in location T
C
      IJ = I + INT((J-I)*R)
      T = IX(IJ)
      TY = IY(IJ)
C
C     If first element of array is greater than T, interchange with T
C
      IF (IX(I) .GT. T) THEN
         IX(IJ) = IX(I)
         IX(I) = T
         T = IX(IJ)
         IY(IJ) = IY(I)
         IY(I) = TY
         TY = IY(IJ)
      ENDIF
      L = J
C
C     If last element of array is less than T, interchange with T
C
      IF (IX(J) .LT. T) THEN
         IX(IJ) = IX(J)
         IX(J) = T
         T = IX(IJ)
         IY(IJ) = IY(J)
         IY(J) = TY
         TY = IY(IJ)
C
C        If first element of array is greater than T, interchange with T
C
         IF (IX(I) .GT. T) THEN
            IX(IJ) = IX(I)
            IX(I) = T
            T = IX(IJ)
            IY(IJ) = IY(I)
            IY(I) = TY
            TY = IY(IJ)
         ENDIF
      ENDIF
C
C     Find an element in the second half of the array which is smaller
C     than T
C
  130 L = L-1
      IF (IX(L) .GT. T) GO TO 130
C
C     Find an element in the first half of the array which is greater
C     than T
C
  140 K = K+1
      IF (IX(K) .LT. T) GO TO 140
C
C     Interchange these elements
C
      IF (K .LE. L) THEN
         TT = IX(L)
         IX(L) = IX(K)
         IX(K) = TT
         TTY = IY(L)
         IY(L) = IY(K)
         IY(K) = TTY
         GO TO 130
      ENDIF
C
C     Save upper and lower subscripts of the array yet to be sorted
C
      IF (L-I .GT. J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 160
C
C     Begin again on another portion of the unsorted array
C
  150 M = M-1
      IF (M .EQ. 0) GO TO 190
      I = IL(M)
      J = IU(M)
C
  160 IF (J-I .GE. 1) GO TO 120
      IF (I .EQ. 1) GO TO 110
      I = I-1
C
  170 I = I+1
      IF (I .EQ. J) GO TO 150
      T = IX(I+1)
      TY = IY(I+1)
      IF (IX(I) .LE. T) GO TO 170
      K = I
C
  180 IX(K+1) = IX(K)
      IY(K+1) = IY(K)
      K = K-1
      IF (T .LT. IX(K)) GO TO 180
      IX(K+1) = T
      IY(K+1) = TY
      GO TO 170
C
C     Clean up
C
  190 IF (KFLAG .LE. -1) THEN
         DO 200 I=1,NN
            IX(I) = -IX(I)
  200    CONTINUE
      ENDIF
      RETURN
      END


      

      
         
